import os
import sys
import toml
import pytest
import contextlib
from setuptools.dist import check_requirements, check_specifier, check_extras

requires = {'python_version': '{0.major}.{0.minor}'.format(sys.version_info)}
source = [{'name': 'pypi', 'url': 'https://pypi.org/simple', 'verify_ssl': True}]


@pytest.fixture
def pf(tmp_path):
    return tmp_path / 'Pipfile'


@pytest.fixture
def pp(tmp_path):
    return tmp_path / 'pyproject.toml'


@contextlib.contextmanager
def chdir(target):
    try:
        start = os.getcwd()
        os.chdir(target)
        yield

    finally:
        os.chdir(start)


@pytest.fixture
def cd_tmp_path(tmp_path):
    return chdir(str(tmp_path.absolute()))


class Spec(dict, toml.decoder.InlineTableDict):
    pass


def write_toml(p, data):
    spec = {
        'source': source,
        'packages': {},
        'dev-packages': {},
        'requires': requires
    }
    spec.update(data)
    p.write_text(toml.dumps(spec, toml.TomlPreserveInlineDictEncoder()))


def write_pyproject(p, data):
    spec = {
        'build-system': {
            'requires': ["setuptools", "wheel"],
            'build-backend': "setuptools.build_meta",
        },
        'tool': {'setuptools-pipfile': data},
    }
    p.write_text(toml.dumps(spec, toml.TomlPreserveInlineDictEncoder()))


class Dist(dict):
    def __setattr__(self, key, value):
        if value:
            if key == 'python_requires':
                check_specifier(self, key, value)
            elif key in ('install_requires', 'tests_require'):
                check_requirements(self, key, value)
            elif key == 'extras_require':
                check_extras(self, key, value)
        self[key] = value


@pytest.fixture
def dist():
    d = Dist()
    return d



import toml
from collections.abc import Mapping
from pathlib import Path
from .pipfile import Pipfile


def use_pipfile(dist, attr, value):
    if not value:
        return

    path = 'Pipfile'
    interpolate = True
    extras = False
    pythons = False
    indexes = None

    if isinstance(value, Mapping):
        path = value.get('path', path)
        interpolate = value.get('interpolate', interpolate)
        extras = value.get('extras', extras)
        pythons = value.get('pythons', pythons)
        indexes = value.get('indexes', indexes)
    elif isinstance(value, (str, Path)):
        path = value
    elif isinstance(value, int):
        extras = value
    elif hasattr(value, '__fspath__'):
        path = value

    pipfile = Pipfile(path, interpolate, pythons, extras, indexes)

    for k, v in pipfile.setup_kwargs().items():
        setattr(dist, k, v)


def finalize_dist(dist, path='pyproject.toml'):
    path = Path(path)
    if not (path.exists() and path.is_file()):
        return

    tools = toml.load(path).get('tool', {})
    value = tools.get('setuptools-pipfile')
    if not value:
        value = 'setuptools-pipfile' in tools

    use_pipfile(dist, None, value)



from setuptools_pipfile.keyword import use_pipfile
from conftest import Spec, write_toml, requires, source


def test_default(pf, dist):
    write_toml(pf, {
        'packages': {'requests': '*'},
        'dev-packages': {'setuptools': '>=41.0.0'},
    })
    use_pipfile(dist, 'use_pipfile', pf)
    assert dist == {
        'install_requires': ['requests'],
        'tests_require': ['setuptools>=41.0.0']
    }


def test_extras_style_one(pf, dist):
    write_toml(pf, {
        'extra': {
            'socks': {
                'requests': Spec({'extras': ['socks'], 'version': '*'}),
            },
        },
    })
    use_pipfile(dist, 'use_pipfile', {'path': pf, 'extras': True})
    assert dist == {
        'extras_require': {
            'socks': ['requests[socks]'],
        },
    }


def test_extras_style_two(pf, dist):
    write_toml(pf, {
        'extra': [
            {
                'name': 'socks',
                'packages': {
                    'requests': Spec({'extras': ['socks'], 'version': '*'}),
                }
            }
        ]
    })
    use_pipfile(dist, 'use_pipfile', {'path': pf, 'extras': {'style': 2}})
    assert dist == {
        'extras_require': {
            'socks': ['requests[socks]'],
        }
    }


def test_extras_style_three(pf, dist):
    write_toml(pf, {
        'dev-packages': {
            'coverage': Spec({'extras': ['toml'], 'version': '*'}),
        }
    })
    use_pipfile(dist, 'use_pipfile', {'path': pf, 'extras': 3})
    assert dist == {
        'extras_require': {
            'dev': ['coverage[toml]'],
        },
        'tests_require': ['coverage[toml]'],
    }


def test_python_requires(pf, dist):
    write_toml(pf, {})
    use_pipfile(dist, 'use_pipfile', {'path': pf, 'pythons': True})
    assert dist == {
        'python_requires': '=={0}'.format(requires['python_version'])
    }


def test_complex_dependencies(pf, dist):
    write_toml(pf, {
        "source": source + [
            {'name': 'test', 'url': 'https://test.pypi.org/simple', 'verify_ssl': True}
        ],
        "dev-packages": {
            "records": ">0.5.0",
            "requests": Spec({
                "version": "*",
                "extras": [
                    "socks"
                ]
            }),
            "django": Spec({
                "git": "https://github.com/django/django.git",
                "ref": "1.11.4",
                "editable": True
            }),
            "e682b37": Spec({
                "file": "https://github.com/divio/django-cms/archive/release/3.4.x.zip"
            }),
            "e1839a8": Spec({
                "path": ".",
                "editable": True
            }),
            "pywinusb": Spec({
                "version": "*",
                "sys_platform": "== 'win32'",
                "index": "test"
            }),
            "unittest2": Spec({
                'version': ">=1.0,<3.0",
                'markers': "python_version < '2.7.9' or (python_version >= '3.0' and python_version < '3.4')",
                'os_name': "== 'nt'"
            })
        }
    })
    use_pipfile(dist, 'use_pipfile', pf)
    assert dist == {
        'tests_require': [
            'django @ git+https://github.com/django/django.git@1.11.4#egg=django',
            'e682b37 @ https://github.com/divio/django-cms/archive/release/3.4.x.zip',
            "pywinusb ; sys_platform == 'win32'",
            'records>0.5.0',
            'requests[socks]',
            "unittest2>=1.0,<3.0 ; (python_version < '2.7.9' or (python_version >= '3.0' and python_version < '3.4')) and os_name == 'nt'",
        ],
        'dependency_links': ['https://test.pypi.org/simple/pywinusb/']
    }


def test_use_pipfile_unset(dist):
    use_pipfile(dist, 'use_pipfile', None)
    assert dist == {}


def test_true(pf, dist, cd_tmp_path):
    write_toml(pf, {
        'packages': {'requests': '*',},
        'dev-packages': {'setuptools': '>=41.0.0',},
    })
    with cd_tmp_path:
        use_pipfile(dist, 'use_pipfile', True)
    assert dist == {
        'install_requires': ['requests'],
        'tests_require': ['setuptools>=41.0.0']
    }


def test_empty_dict(pf, dist, cd_tmp_path):
    write_toml(pf, {
        'packages': {'requests': '*',},
        'dev-packages': {'setuptools': '>=41.0.0',},
    })
    with cd_tmp_path:
        use_pipfile(dist, 'use_pipfile', {})
    assert dist == {}


def test_integer_one(pf, dist, cd_tmp_path):
    write_toml(pf, {
        'extra': {
            'socks': {
                'requests': Spec({'extras': ['socks'], 'version': '*'}),
            },
        },
    })
    with cd_tmp_path:
        use_pipfile(dist, 'use_pipfile', 1)
    assert dist == {
        'extras_require': {
            'socks': ['requests[socks]'],
        },
    }


def test_integer_two(pf, dist, cd_tmp_path):
    write_toml(pf, {
        'extra': [
            {
                'name': 'socks',
                'packages': {
                    'requests': Spec({'extras': ['socks'], 'version': '*'}),
                }
            }
        ]
    })
    with cd_tmp_path:
        use_pipfile(dist, 'use_pipfile', 2)
    assert dist == {
        'extras_require': {
            'socks': ['requests[socks]'],
        }
    }


def test_integer_three(pf, dist, cd_tmp_path):
    write_toml(pf, {
        'dev-packages': {
            'coverage': Spec({'extras': ['toml'], 'version': '*'}),
        }
    })
    with cd_tmp_path:
        use_pipfile(dist, 'use_pipfile', 3)
    assert dist == {
        'extras_require': {
            'dev': ['coverage[toml]'],
        },
        'tests_require': ['coverage[toml]'],
    }
